# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import os
import pickle
import concurrent.futures
from concurrent.futures import ThreadPoolExecutor

import numpy as np


class LightField(object):
    """container for accessing sampled data

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    rebuild: bool, optional
        build kd-tree even if one exists
    prefix: str, optional
        prefix of data files to map
    """

    def __init__(self, scene, rebuild=False, prefix='sky', srcn=1, rmraw=False):
        #: bool: force rebuild kd-tree
        self.rebuild = rebuild
        self.srcn = srcn
        #: str: prefix of data files from sampler (stype)
        self.prefix = prefix
        self._vec = None
        self._lum = None
        self._omega = None
        self._rmraw = rmraw
        self.scene = scene
        self._rawfiles = self.raw_files()

    def __del__(self):
        try:
            if self._rmraw:
                for rf in self._rawfiles:
                    try:
                        os.remove(rf)
                    except IOError:
                        pass
        except AttributeError:
            pass

    def raw_files(self):
        return []

    @property
    def vec(self):
        """direction vector (3,)"""
        return self._vec

    @property
    def lum(self):
        """luminance (srcn,)"""
        return self._lum

    @property
    def omega(self):
        """solid angle (1,)"""
        return self._omega

    def outfile(self, idx):
        istr = "_".join([f"{i:04d}" for i in np.asarray(idx).reshape(-1)])
        return f"{self.scene.outdir}_{self.prefix}_{istr}"

    def items(self):
        return range(self.scene.area.npts)





# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

import os
import re
import shutil
import configparser
import sys

import numpy as np
import json

from clasp import script_tools as cst
from clasp.click_callbacks import parse_file_list
from scipy.spatial import cKDTree

from raytraverse.mapper import SpaceMapper, ViewMapper, SpaceMapperPt


class Scene(object):
    """container for scene description

    Parameters
    ----------
    outdir: str
        path to store scene info and output files
    scene: str, optional (required if not reload)
        space separated list of radiance scene files (no sky) or octree
    area: str, optional (required if not reload)
        radiance scene file containing planar geometry of analysis area
        or a list of points (line per point, space seperated, first 3 columns
        x, y, z
    reload: bool, optional
        if True attempts to load existing scene files in new instance
        overrides 'overwrite'
    overwrite: bool, optional
        if True and outdir exists, will overwrite, else raises a FileExistsError
    ptres: float, optional
        final spatial resolution in scene geometry units
    ptro: float, optional
        angle in degrees counter-clockwise to point grid
    pttol: float, optional
        tolerance for point search when using point list for area
    viewdir: (float, float, float), optional
        vector (x,y,z) view direction (orients UV space)
    viewangle: float, optional
        should be 1-180 or 360
    skyres: float, optional
        approximate square patch size (sets sun resolution too)
    maxspec: float, optional
        maximum specular transmission in scene
        (used to clip pdf for sun sampling)
    """

    def __init__(self, outdir, scene=None, area=None, reload=True,
                 overwrite=False, ptres=1.0, ptro=0.0, pttol=1.0,
                 viewdir=(0, 1, 0), viewangle=360, skyres=10.0, maxspec=0.3,
                 **kwargs):
        locvar = locals()
        try:
            os.mkdir(outdir)
        except FileExistsError as e:
            if overwrite:
                shutil.rmtree(outdir)
                os.mkdir(outdir)
            elif reload:
                pass
            else:
                raise e
        js = f'{outdir}/scene_parameters.json'
        if os.path.isfile(js):
            with open(js, 'r') as jf:
                params = json.load(jf)
            os.remove(js)
            print(f'Scene parameters loaded from {js}', file=sys.stderr)
            self.__init__(**params)
        else:
            locvar.pop('self')
            locvar.pop('kwargs')
            with open(js, 'w') as jf:
                json.dump(locvar, jf)
            #: bool: try to reload scene files
            self.reload = reload
            #: str: path to store scene info and output files
            self.outdir = outdir
            a = f'{self.outdir}/area.rad'
            if self.reload and os.path.isfile(a):
                pass
            else:
                try:
                    shutil.copy(area, a)
                except TypeError:
                    raise ValueError('Cannot initialize Scene with '
                                     f'area={area}')
            try:
                ptload = np.loadtxt(a)[:, 0:3]
            except ValueError:
                self.area = SpaceMapper(a, ptres, ptro, pttol)
            else:
                self.area = SpaceMapperPt(a, ptres, ptro, pttol)
            #: float: maximum specular transmission in scene
            self.maxspec = maxspec
            self._solarbounds = None
            self.scene = scene
            self.reload = False
            #: raytraverse.viewmapper.ViewMapper: view translation class
            self.view = ViewMapper(viewdir, viewangle)
            if skyres < .7:
                print('Warning! minimum sunres is .7 to avoid overlap and',
                      file=sys.stderr)
                print('allow for jittering position, sunres set to .7',
                      file=sys.stderr)
                skyres = .7
            self.skyres = skyres
            self.pt_kd = None

    @property
    def skyres(self):
        return self._skyres

    @skyres.setter
    def skyres(self, s):
        self._skyres = int(np.floor(90/s)*2)

    @property
    def scene(self):
        """render scene files (octree)

        :getter: Returns this samplers's scene file path
        :setter: Sets this samplers's scene file path and creates run files
        :type: str
        """
        return self._scene

    @scene.setter
    def scene(self, scene):
        o = f'{self.outdir}/scene.oct'
        if self.reload and os.path.isfile(o):
            pass
        else:
            dims = cst.pipeline([f'getinfo -d {scene}', ])
            try:
                m = re.match(scene + r': [\d.-]+ [\d.-]+ [\d.-]+ [\d.-]+',
                             dims.strip())
            except TypeError:
                raise ValueError(f'{o} does not exist, Scene() must be invoked'
                                 ' with a scene= argument')
            if m:
                oconv = f'oconv -i {scene}'
            else:
                scene = " ".join(parse_file_list(None, scene))
                oconv = f'oconv -f {scene}'
            result, err = cst.pipeline([oconv, ],
                                       outfile=o,
                                       close=True, caperr=True, writemode='wb')
            if b'fatal' in err:
                raise ChildProcessError(err.decode(cst.encoding))
        self._scene = o

    def pts(self):
        return self.area.pts()




# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
from raytraverse.renderer.radiancerenderer import RadianceRenderer
from raytraverse.crenderer import cRtrace


class Rtrace(RadianceRenderer):
    """singleton wrapper for c++ crenderer.cRtrace singleton class"""
    Engine = cRtrace
    name = 'rtrace'

    @classmethod
    def update_ospec(cls, vs, of='a'):
        if not cls.initialized:
            raise ValueError('Rtrace instance not initialized')
        cls.instance.update_ospec(vs, of)

    @classmethod
    def new_amb(cls, af):
        cls.instance.new_amb(af)

    @classmethod
    def load_source(cls, srcname, freesrc=-1):
        cls.instance.load_source(srcname, freesrc)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

import numpy as np

from raytraverse import translate, skycalc


class SkyInfo(object):
    """sky location data object

    Parameters
    ----------
    loc: tuple
        lat, lon, tz (in degrees, west is positive
    skyro: float
        sky rotation (in degrees, ccw)
    """

    def __init__(self, loc, skyro=0.0):
        #: float: ccw rotation (in degrees) for sky
        self.skyro = skyro
        self.loc = loc

    @property
    def solarbounds(self):
        """read only extent of solar bounds for given location
        set via loc

        :getter: Returns solar bounds
        :type: (np.array, np.array)
        """
        return self._solarbounds

    @property
    def loc(self):
        """scene location

        :getter: Returns location
        :setter: Sets location and self.solarbounds
        :type: (float, float, int)
        """
        return self._loc

    @loc.setter
    def loc(self, loc):
        """
        generate UV coordinates for jun 21 and dec 21 to use for masking
        sky positions
        """
        self._loc = loc
        if loc is None:
            self._solarbounds = None
        else:
            jun = np.arange('2020-06-21', '2020-06-22', 5,
                            dtype='datetime64[m]')
            dec = np.arange('2020-12-21', '2020-12-22', 5,
                            dtype='datetime64[m]')
            jxyz = skycalc.sunpos_xyz(jun, *loc)
            dxyz = skycalc.sunpos_xyz(dec, *loc)
            juv = translate.xyz2uv(jxyz[jxyz[:, 2] > 0], flipu=False)
            duv = translate.xyz2uv(dxyz[dxyz[:, 2] > 0], flipu=False)
            juv = juv[juv[:, 0].argsort()]
            duv = duv[duv[:, 0].argsort()]
            self._solarbounds = (juv, duv)

    def in_solarbounds(self, uv, size=0.0):
        """
        for checking if src direction is in solar transit

        Parameters
        ----------
        uv: np.array
            source directions
        size: float
            offset around UV to test

        Returns
        -------
        result: np.array
            Truth of ray.src within solar transit
        """
        if self.loc is None:
            return np.full(uv.shape[0], True)
        if abs(self.skyro) > 0:
            xyz = translate.uv2xyz(uv, xsign=1)
            rxyz = translate.rotate_elem(xyz, -self.skyro)
            uv = translate.xyz2uv(rxyz, flipu=False)
        o = size/2
        juv, duv = self.solarbounds
        vlowleft = duv[np.searchsorted(duv[:, 0], uv[:, 0] - o) - 1]
        vlowright = duv[np.searchsorted(duv[:, 0], uv[:, 0] - o) - 1]
        vupleft = juv[np.searchsorted(juv[:, 0], uv[:, 0] + o) - 1]
        vupright = juv[np.searchsorted(juv[:, 0], uv[:, 0] + o) - 1]
        inbounds = np.stack((vlowleft[:, 1] <= uv[:, 1] - o,
                             vlowright[:, 1] <= uv[:, 1] - o,
                             uv[:, 1] + o <= vupleft[:, 1],
                             uv[:, 1] + o <= vupright[:, 1]))
        return np.all(inbounds, 0)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

# c++ compiling and pybind setup code copied from https://github.com/pybind/python_example.git

# License:

# Copyright (c) 2016 The Pybind Development Team, All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
#    may be used to endorse or promote products derived from this software
#    without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# You are under no obligation whatsoever to provide any bug fixes, patches, or
# upgrades to the features, functionality or performance of the source code
# ("Enhancements") to anyone; however, if you choose to make your Enhancements
# available either publicly, or directly to the author of this software, without
# imposing a separate written license agreement for such Enhancements, then you
# hereby grant the following license: a non-exclusive, royalty-free perpetual
# license to install, use, modify, prepare derivative works, incorporate into
# other computer software, distribute, and sublicense such enhancements or
# derivative works thereof, in binary and source code form.


"""The setup script."""
from setuptools import find_packages
from skbuild import setup
# from setuptools.command.build_ext import build_ext
# from distutils import ccompiler
# import sys
# import setuptools

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = ['clasp', 'numpy', 'scipy', 'pywavelets', 'matplotlib',
                'skyfield', 'clipt', 'pybind11']

setup_requirements = ["setuptools", "wheel", "scikit-build", "cmake", "ninja"]

test_requirements = ['pytest', 'hdrstats', 'pytest-cov']

# TODO: add necessary radiance .cal files (and make sure source files are
# included in in dist before creating pypi.
data_files = []
package_data = {}

setup(
    author="Stephen Wasilewski",
    author_email='stephanwaz@gmail.com',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="variance based adaptive sampling for CBDM",
    python_requires=">=3.6",
    entry_points={
        'console_scripts': [
            'raytraverse=raytraverse.cli:main',
            'genskyvec_sc=raytraverse.gsv:main'
        ],
    },
    cmake_minimum_required_version="3.15",
    install_requires=requirements,
    license="Mozilla Public License 2.0 (MPL 2.0)",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='raytraverse',
    name='raytraverse',
    packages=find_packages(),
    data_files=data_files,
    package_data=package_data,
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://gitlab.enterpriselab.ch/lightfields/raytraverse',
    version='0.1.1',
    zip_safe=False,
    )



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

import os

import numpy as np

from raytraverse import translate, renderer
from raytraverse.sampler.sampler import Sampler


class SCBinSampler(Sampler):
    """sample contributions from the sky hemisphere according to a square grid
    transformed by shirley-chiu mapping using rcontrib.

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    srcn: int, optional
        side of square sky resolution
    """

    def __init__(self, scene, accuracy=1,
                 rcopts='-ab 7 -ad 60000 -as 30000 -lw 1e-7', **kwargs):
        f = open(f'{scene.outdir}/scbins.cal', 'w')
        f.write(translate.scbinscal)
        f.close()
        skydeg = ("void glow skyglow 0 0 4 1 1 1 0 skyglow source sky 0 0 4"
                  " 0 0 1 180")
        self.engine = renderer.Rcontrib()
        self.engine.reset()
        srcn = scene.skyres**2
        engine_args = (f"-V+ {rcopts} -Z+ -e 'side:{scene.skyres}' -f "
                       f"{scene.outdir}/scbins.cal -b bin -bn {srcn} "
                       f"-m skyglow")
        super().__init__(scene, srcn=srcn, stype='sky',  srcdef=skydeg,
                         accuracy=accuracy, engine_args=engine_args, **kwargs)

    def __del__(self):
        super().__del__()
        try:
            os.remove(f'{self.scene.outdir}/scbins.cal')
        except (IOError, TypeError):
            pass

    def sample(self, vecf):
        """call rendering engine to sample sky contribution

        Parameters
        ----------
        vecf: str
            path of file name with sample vectors
            shape (N, 6) vectors in binary float format

        Returns
        -------
        lum: np.array
            array of shape (N,) to update weights
        """
        lum = super().sample(vecf)
        return np.max(lum, 1)



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""functions for plotting data"""
import sys

import numpy as np
import matplotlib.pyplot as plt
from clipt import mplt


def imshow(im, figsize=(10, 10), outf=None, **kwargs):
    fig, ax = plt.subplots(1, 1, figsize=figsize)
    ax.imshow(im.T, origin='lower', **kwargs)
    ax.set_xticks([])
    ax.set_yticks([])
    plt.tight_layout()
    if outf is None:
        plt.show()
    else:
        plt.savefig(outf)
    plt.close(fig)


def hist(lums, bins='auto', outf=None, **kwargs):
    h, binedges = np.histogram(lums.ravel(), bins=bins, **kwargs)
    print(h, binedges, file=sys.stderr)
    b = np.repeat(binedges, 2)[1:-1]
    h = np.repeat(h, 2)
    mplt.quick_scatter([b], [h], outf=outf)



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import os
import shlex
from subprocess import Popen, PIPE
from raytraverse import io
from raytraverse.renderer.renderer import Renderer


class SPRenderer(Renderer):
    """Subprocess renderer class"""
    cleanup = "rcalc -if3 -of -e $1=.265074126*$1+.670114631*$2+.064811243*$3"

    @classmethod
    def initialize(cls, args, scene, nproc=None, iot="ff"):
        if nproc is None:
            nproc = os.cpu_count()
        if args is not None and not cls.initialized:
            if iot[-1] == 'a':
                raise ValueError(f'{cls.__name__} must output binary format')
            cls.cleanup = (f"rcalc -i{iot[-1]}3 -o{iot[-1]}"
                           " -e $1=.265074126*$1+.670114631*$2+.064811243*$3")
            cls.initialized = cls._set_args(args + " -h- " + scene, iot, nproc)
            # TODO: populate header
            cls.header = ""

    @classmethod
    def call(cls, rayfile, store=True, outf=None, vecs2stdin=True):
        if not cls.initialized:
            raise ValueError(f'{cls.__name__} instance not initialized')
        if cls.Engine is None:
            raise NotImplementedError(f'{cls.__name__} is a virtual class')
        with io.CaptureStdOut(True, store, outf) as capture:
            if vecs2stdin:
                p = Popen(cls.initialized, stdout=PIPE,
                          stdin=open(rayfile, 'rb'))
            else:
                p = Popen(cls.initialized + [rayfile], stdout=PIPE)
            q = Popen(shlex.split(cls.cleanup), stdin=p.stdout)
            q.communicate()
        return capture.stdout


class SPRtrace(SPRenderer):
    Engine = 'rtrace'
    name = 'rtrace'


class SPRcontrib(SPRenderer):
    Engine = 'rcontrib'
    name = 'rcontrib'




# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
import os
import sys

import numpy as np

import clasp.script_tools as cst
from raytraverse import translate, io, draw, quickplot


class Sampler(object):
    """base sampling class

    To implement a proper-subclass an engine attribute must be set to a
    renderer instance prior to calling Sampler.__init__. Also, the method
    sample must be overriden to properly set up arguments for the renderer.call

    Parameters
    ----------
    scene: raytraverse.scene.Scene
        scene class containing geometry, location and analysis plane
    fdres: int, optional
        final directional resolution given as log2(res)
    srcn: int, optional
        number of sources return per vector by run
    accuracy: float, optional
        parameter to set threshold at sampling level relative to final level
        threshold (smaller number will increase sampling, default is 1.0)
    idres: int, optional
        initial direction resolution (as log2(res))
    stype: str, optional
        sampler type (prefixes output files)
    srcdef: str, optional
        path or string with source definition to add to scene
    plotp: bool, optional
        show probability distribution plots at each level (first point only)
    bands: int, optional
        number of spectral bands returned by the engine
    engine_args: str, optional
        command line arguments used to initialize engine
    nproc: int, optional
        number of processors to give to the engine, if None, uses os.cpu_count()
    """
    engine = None

    def __init__(self, scene, fdres=9, srcn=1, accuracy=1.0, idres=4,
                 stype='generic', srcdef=None, plotp=False,
                 bands=1, engine_args="", nproc=None, **kwargs):
        name = type(self).__name__
        self._staticscene = True
        if name == "Sampler":
            raise NotImplementedError("Sampler base class should not be "
                                      "instantiated directly")
        elif not all(hasattr(self.engine, atr) for atr
                     in ['initialize', 'call']):
            raise NotImplementedError(f"Subclass {name} of Sampler is ill"
                                      " defined, an engine attribute must"
                                      " be set to a valid renderer class"
                                      " before invoking Sampler.__init__")
        #: int: number of spectral bands / channels returned by renderer
        #: based on given renderopts (user ensures these agree).
        self.bands = bands
        self.scene = scene
        #: func: mapper to use for sampling
        self.samplemap = self.scene.view
        #: int: number of sources return per vector by run
        self.srcn = srcn
        self.accuracy = accuracy
        #: int: initial direction resolution (as log2(res))
        self.idres = idres
        self.levels = fdres
        #: np.array: holds weights for self.draw
        self.weights = np.full(np.concatenate((self.scene.area.ptshape,
                                               self.levels[0])), 1e-7,
                               dtype=np.float32)
        #: int: index of next level to sample
        self.idx = 0
        #: str: sampler type
        self.stype = stype
        self.compiledscene = srcdef
        self.engine.initialize(engine_args, self.compiledscene, nproc=nproc,
                               iot="ff")
        self.plotp = plotp
        self.levelsamples = np.ones(self.levels.shape[0])
        # track vector files written for concatenation / cleanup after run
        self._vecfiles = []

    def __del__(self):
        if not self._staticscene:
            try:
                os.remove(self.compiledscene)
            except (IOError, TypeError):
                pass

    @property
    def compiledscene(self):
        return self._compiledscene

    @compiledscene.setter
    def compiledscene(self, src):
        self._staticscene = src is None
        if self._staticscene:
            self._compiledscene = f'{self.scene.outdir}/scene.oct'
        else:
            self._compiledscene = f'{self.scene.outdir}/{self.stype}.oct'
            if os.path.isfile(src):
                ocom = f'oconv -f -i {self.scene.outdir}/scene.oct {src}'
                inp = None
            else:
                ocom = f'oconv -f -i {self.scene.outdir}/scene.oct -'
                inp = src
            f = open(self.compiledscene, 'wb')
            cst.pipeline([ocom], outfile=f, inp=inp, close=True)

    @property
    def idx(self):
        """sampling level

        :getter: Returns the sampling level
        :setter: Set the sampling level and associated values (temp, shape)
        :type: int
        """
        return self._idx

    @idx.setter
    def idx(self, idx):
        self._idx = idx

    @property
    def levels(self):
        """sampling scheme

        :getter: Returns the sampling scheme
        :setter: Set the sampling scheme from (ptres, fdres, skres)
        :type: np.array
        """
        return self._levels

    @levels.setter
    def levels(self, fdres):
        """calculate sampling scheme"""
        self._levels = np.array([(2**i*self.scene.view.aspect, 2**i)
                                 for i in range(self.idres, fdres+1, 1)])

    @property
    def scene(self):
        """scene information

        :getter: Returns this sampler's scene
        :setter: Set this sampler's scene and create octree with source desc
        :type: raytraverse.scene.Scene
        """
        return self._scene

    @scene.setter
    def scene(self, scene):
        """Set this sampler's scene and create sky octree"""
        self._scene = scene

    def sample(self, vecf):
        """generic sample function
        """
        outf = f'{self.scene.outdir}/{self.stype}_vals.out'
        f = open(outf, 'a+b')
        lumb = self.engine.call(vecf, outf=f)
        f.close()
        shape = (-1, self.srcn, self.bands)
        lum = io.bytes2np(lumb, shape)
        return np.squeeze(lum, axis=-1)

    def _uv2xyz(self, uv, si):
        """including to allow overriding mapping bevahior of daughter classes"""
        return self.samplemap.uv2xyz(uv)

    def sample_idx(self, pdraws):
        """generate samples vectors from flat draw indices

        Parameters
        ----------
        pdraws: np.array
            flat index positions of samples to generate

        Returns
        -------
        si: np.array
            index array of draws matching samps.shape
        vecs: np.array
            sample vectors
        """
        shape = np.concatenate((self.scene.area.ptshape, self.levels[self.idx]))
        # index assignment
        si = np.stack(np.unravel_index(pdraws, shape))
        # convert to UV directions and positions
        uv = si.T[:, -2:]/shape[3]
        pos = self.scene.area.uv2pt((si.T[:, 0:2] + .5)/shape[0:2])
        uv += (np.random.default_rng().random(uv.shape))/shape[3]
        # mplt.quick_scatter([uv[:, 0]], [uv[:, 1]], ms=3, lw=0)
        if pos.size == 0:
            xyz = pos
        else:
            xyz = self._uv2xyz(uv, si)
        vecs = np.hstack((pos, xyz))
        return si, vecs

    def dump_vecs(self, vecs, si=None):
        """save vectors to file

        Parameters
        ----------
        vecs: np.array
            ray directions to write
        si: np.array, optional
            sample indices
        """
        outf = f'{self.scene.outdir}/{self.stype}_vecs_{self.idx:02d}.out'
        f = open(outf, 'wb')
        f.write(io.np2bytes(vecs))
        f.close()
        if si is not None:
            ptidx = np.ravel_multi_index((si[0], si[1]), self.scene.area.ptshape)
            outif = f'{self.scene.outdir}/{self.stype}_vecpidx_{self.idx:02d}.out'
            f = open(outif, 'wb')
            f.write(io.np2bytes(ptidx))
            f.close()
        else:
            outif = None
        self._vecfiles.append((outif, outf))
        return outf

    def draw(self):
        """draw samples based on detail calculated from weights
        detail is calculated across direction only as it is the most precise
        dimension

        Returns
        -------
        pdraws: np.array
            index array of flattened samples chosen to sample at next level
        """
        dres = self.levels[self.idx]
        pres = self.scene.area.ptshape
        if self.idx == 0 and np.var(self.weights) < 1e-9:
            pdraws = np.arange(np.prod(dres)*np.prod(pres))
        else:
            # direction detail
            daxes = (len(pres) + len(dres) - 2, len(pres) + len(dres) - 1)
            p = draw.get_detail(self.weights, daxes)
            if self.plotp:
                quickplot.imshow(np.log10(p.reshape(self.weights.shape)[0, 0]),
                                 [20, 10])
            # draw on pdf
            # threshold is set to accurracy at final
            threshold = self.accuracy * 4**(self.idx - len(self.levels))
            pdraws = draw.from_pdf(p, threshold)
        return pdraws

    def update_pdf(self, si, lum):
        """update self.weights (which holds values used to calculate pdf)

        Parameters
        ----------
        si: np.array
            multidimensional indices to update
        lum:
            values to update with

        """
        self.weights[tuple(si)] = lum

    def run_callback(self):
        outf = f'{self.scene.outdir}/{self.stype}_vecs.out'
        f = open(outf, 'wb')
        for idxf, vecf in self._vecfiles:
            fsrc = open(vecf, 'rb')
            vecs = io.bytefile2np(fsrc, (-1, 6))
            fsrc.close()
            os.remove(vecf)
            if idxf is not None:
                fidx = open(idxf, 'rb')
                ptidx = io.bytefile2np(fidx, (1, -1))
                fidx.close()
                os.remove(idxf)
                f.write(io.np2bytes(np.vstack((ptidx.reshape(1, -1),
                                               vecs[:, 3:].T)).T))
            else:
                f.write(io.np2bytes(vecs))
        f.close()

    def get_scheme(self):
        scheme = np.ones((self.levels.shape[0], self.levels.shape[1] + 4))
        scheme[:, 2:-2] = self.levels
        scheme[:, 0:2] = self.scene.area.ptshape
        scheme[:, -2] = self.srcn
        scheme[:, -1] = self.levelsamples
        return scheme.astype(int)

    # @profile
    def run(self):
        """execute sampler"""

        allc = 0
        f = open(f'{self.scene.outdir}/{self.stype}_vals.out', 'wb')
        f.close()
        print('Sampling...', file=sys.stderr)
        hdr = ['level', 'shape', 'samples', 'rate', 'filesize (MB)']
        print('{:>8}  {:>25}  {:<10}  {:<8}  {}'.format(*hdr), file=sys.stderr)
        fsize = 0
        for i in range(self.idx, self.levels.shape[0]):
            shape = np.concatenate((self.scene.area.ptshape, self.levels[i]))
            self.idx = i
            self.weights = translate.resample(self.weights, shape)
            draws = self.draw()
            if draws is None:
                srate = 0.0
                row = [f'{i + 1} of {self.levels.shape[0]}', str(shape),
                       0, f"{srate:.02%}", fsize]
                print('{:>8}  {:>25}  {:<10}  {:<8}  {:.03f}'.format(*row),
                      file=sys.stderr)
            else:
                self.levelsamples[self.idx] = draws.size
                si, vecs = self.sample_idx(draws)
                srate = si.shape[1]/np.prod(shape)
                fsize += 4*self.bands*self.srcn*si.shape[1]/1000000
                row = [f'{i+1} of {self.levels.shape[0]}', str(shape),
                       si.shape[1], f"{srate:.02%}", fsize]
                print('{:>8}  {:>25}  {:<10}  {:<8}  {:.03f}'.format(*row),
                      file=sys.stderr)
                vecf = self.dump_vecs(vecs, si)
                lum = self.sample(vecf)
                self.update_pdf(si, lum)
                a = lum.shape[0]
                allc += a
        print("-"*70, file=sys.stderr)
        srate = allc/self.weights.size
        row = ['total sampling:', allc, f"{srate:.02%}", fsize]
        print('{:<35}  {:<10}  {:<8}  {:.03f}'.format(*row), file=sys.stderr)
        self.run_callback()



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""functions for plotting data"""
import warnings

import numpy as np
from clipt.plot import get_colors
from matplotlib import rcParams
from matplotlib.figure import Figure
from matplotlib.colors import Normalize
from matplotlib.patches import Polygon


def save_img(fig, ax, outf, title=None):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        fig.tight_layout()
        if title is not None:
            ax.set_title(title)
        fig.savefig(outf)


def imshow(im, figsize=(10, 10), outf=None, **kwargs):
    fig = Figure(figsize=figsize)
    ax = fig.add_subplot()
    ax.imshow(im.T, origin='lower', **kwargs)
    ax.set_xticks([])
    ax.set_yticks([])
    save_img(fig, ax, outf=outf)


def mk_img_setup(lums, bounds=None, figsize=(10, 10), ext=1):
    defparam = {
        'font.weight': 'ultralight',
        'font.family': 'sans-serif',
        'font.size': 15,
        'axes.linewidth': 0,
        }
    rcParams.update(defparam)
    if bounds is None:
        minl = np.min(lums)
        maxl = np.max(lums)
    else:
        minl = bounds[0]
        maxl = bounds[1]
        lums = np.clip(lums, minl, maxl)
    fig = Figure(figsize=figsize)
    ax = fig.add_subplot()
    norm = Normalize(vmin=minl, vmax=maxl)
    lev = np.linspace(minl, maxl, 200)
    ext = np.asarray(ext).flatten()
    if ext.size == 1:
        ax.set(xlim=(-ext, ext), ylim=(-ext, ext))
    elif ext.size == 2:
        ax.set(xlim=ext, ylim=ext)
    else:
        ax.set(xlim=ext[0:2], ylim=ext[2:4])
    return lums, fig, ax, norm, lev


def set_ang_ticks(ax, ext):
    ax.tick_params(length=10, width=.5, direction='inout', pad=5)
    ticks = np.linspace(-ext, ext, 7)
    labs = np.round(np.linspace(-ext*180/np.pi,
                                ext*180/np.pi, 7)).astype(int)
    ax.set_yticks(ticks)
    ax.set_yticklabels(labs)
    ax.set_xticks(ticks)
    ax.set_xticklabels(labs)


def colormap(colors, norm):
    cmap = get_colors(colors)
    cmap.set_norm(norm)
    return cmap


def plot_patches(ax, patches, patchargs=None):
    if patchargs is None:
        patchargs = {}
    for lum, p in patches:
        ax.add_patch(Polygon(p, closed=True, facecolor=lum, **patchargs))



# -*- coding: utf-8 -*-

# Copyright (c) 2019 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""standardized metric functions"""
import numpy as np


def illum(vm, vec, omega, lum, scale=179, **kwargs):
    return np.einsum('i,i,i->', vm.ctheta(vec), lum, omega) * scale


def avglum(vm, vec, omega, lum, scale=179, area=None, **kwargs):
    if area is None:
        area = np.sum(omega)
    return np.einsum('i,i->', lum, omega) * scale / area


def sqlum(vm, vec, omega, lum, scale=179, area=None, **kwargs):
    if area is None:
        area = np.sum(omega)
    alum2 = avglum(vm, vec, omega, lum, scale=scale, area=area, **kwargs)**2
    a2lum = np.einsum('i,i,i->', lum, lum, omega) * scale**2 / area
    return a2lum / alum2



# -*- coding: utf-8 -*-
# Copyright (c) 2020 Stephen Wasilewski
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""wavelet and associated probability functions."""
import numpy as np
import pywt
from raytraverse import craytraverse


def get_detail(samps, axes):
    """run high pass filter over given axes"""

    # filterbank with pad and offset slice centers distribution around variance
    wav = pywt.Wavelet('custom4', ([0.40824829, 0.81649658, 0.40824829, 0.],
                                   [-0.40824829, 0.81649658, -0.40824829, 0],
                                   [0, 0.40824829, 0.81649658, 0.40824829],
                                   [-0.40824829, 0.81649658, 0.40824829, 0]))
    # mod adds extra padding to ensure evenness of transformed dimensions
    padding = [(2, 2 + int(np.mod(s, 2))) if i in axes else (0, 0) for i, s in
               enumerate(samps.shape)]
    snn = slice(None, None, None)
    slicing = (snn, ) + tuple([slice(1, -3 - int(np.mod(s, 2)), None)
                               if i in axes else snn for i, s in
                               enumerate(samps.shape)])
    psamps = pywt.pad(samps, padding, mode='symmetric')
    # calculate horiz, vert and diagonal detail
    d = pywt.swtn(psamps, wav, 1, trim_approx=True, axes=axes)
    d = np.asarray(tuple(d[1].values()))[slicing]
    # sum over detail and normalize (useful for non parametric sampling rates)
    # the detail can be read as delta luminance around that pixel
    d_det = np.sum(np.abs(d), 0).ravel() * 3
    m = np.nanmean(d_det)
    return np.where(np.isfinite(d_det), d_det, m)


def from_pdf(pdf, threshold):
    candidates = np.empty(pdf.size, dtype=np.uint32)
    bidx = np.empty(pdf.size, dtype=np.uint32)
    cs, bs, nsampc = craytraverse.from_pdf(pdf, candidates, bidx, threshold)
    if nsampc == 0:
        return bidx[:bs]
    # if normalization happens in c-func floating point precision does not
    # guarantee that pdfc adds to 1, which choice() requires.
    pdfc = pdf[candidates[:cs]]/np.sum(pdf[candidates[:cs]])
    cidx = np.random.default_rng().choice(candidates[:cs], nsampc,
                                          replace=False, p=pdfc)
    return np.concatenate((bidx[:bs], cidx))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for raytraverse.draw"""

import pytest
from raytraverse import draw, craytraverse
import numpy as np
from scipy.stats import norm, hypergeom, halfnorm
from scipy import stats
from clipt import mplt


# def test_version():
#     # with craytraverse.ostream_redirect():
#     #     craytraverse.rtrace("rtrace -defaults".split())
#     print(craytraverse.version())
#
#     # print(craytraverse.__doc__)
#     # print(craytraverse.rtdefaults())


def test_from_pdf():
    rv = stats.norm()
    nsamp = 500000
    t = .01
    x = np.linspace(0, 5, nsamp)
    pdf = rv.pdf(x)
    exp = np.sum(pdf > t)
    # c2 = draw.from_pdf2(pdf, t)
    c = draw.from_pdf(pdf, t)
    # print(np.sum(pdf > t), c.size, c2.size)
    # hist = np.histogram(x[c], 50)
    # mplt.quick_scatter([hist[1][1:], [0, 0]], [hist[0], [0, 7000]])
    # print(exp, c.size, c2.size)
    assert np.isclose(c.size, exp)


